import glob
import logging
import os
import shutil
from pathlib import Path
from typing import Dict, List

from .interfaces import ConveyorIF

"""
Post QA: Retrieve results directories from QA2 area to the spool area 
"""


class RetrievalConveyor(ConveyorIF):
    def __init__(self, settings: Dict[str, str], action: str):
        self.logger = logging.getLogger("conveyor")
        self.action = action
        self.settings = settings

    def retrieval(self):
        qa_path = self.determine_qa_directory()
        spool_path = self.determine_spool_directory()

        os.chdir(spool_path)

        contents = glob.glob("*/", recursive=True)

        self.break_symlinks(spool_path, contents)
        self.move_subdirectories_to_spool(qa_path, spool_path, contents)
        check = self.check_spool_contents(spool_path, contents)
        if check:
            Path(qa_path).rmdir()

    def break_symlinks(self, spool_path: Path, dir_list: List[str]):
        self.logger.info(
            f"Breaking symlinks between qa2 and spool for directory {spool_path.stem}..."
        )
        for directory in dir_list:
            directory = directory.strip("/")
            if Path(spool_path / directory).is_symlink():
                Path(spool_path / directory).unlink()
                self.logger.info(f"Symlink for {directory} removed.")
            else:
                self.logger.info(f"Directory {directory} is not a symlink! Skipping.")

    def move_subdirectories_to_spool(self, qa_path: Path, spool_path: Path, dir_list: List[str]):
        self.logger.info(f"Moving directories from qa2 to spool for directory {qa_path.stem}...")
        for directory in dir_list:
            qa_dir = qa_path / directory
            spool_dir = spool_path / directory
            self.logger.info(f"Moving directory {directory}...")
            shutil.move(qa_dir, spool_dir)
            self.logger.info("Done.")

    def determine_qa_directory(self) -> Path:
        qa_area = self.settings["qa_delivery_area"]
        wf_dir = self.settings["destination_subdir"]
        return Path(qa_area + "/" + wf_dir)

    def determine_spool_directory(self) -> Path:
        spool_area = self.settings["workspaces_lustre_root_dir"]
        wf_dir = self.settings["destination_subdir"]
        return Path(spool_area + "/" + wf_dir)

    def check_spool_contents(self, spool_path: Path, expected_dirs: List[str]) -> bool:
        os.chdir(spool_path)

        contents = glob.glob("*/", recursive=True)
        if set(contents).issubset(expected_dirs) and not len(contents) < len(expected_dirs):
            self.logger.info("All directories were successfully transferred!")
            return True
        else:
            self.logger.error(
                f"ERROR: A directory seems to be missing! "
                f"Directory list {contents} does not match the "
                f"list of expected directories!"
            )
            return False

    def convey(self):
        self.logger.info("RUNNING POST QA STANDARD CALIBRATION DIRECTORY RETRIEVAL!")
        self.retrieval()
