""" Tests for the execution manager """
import logging
import sys
from typing import Dict, List
from unittest.mock import patch

# pylint: disable=E0401, W0212, W0621

import pytest
from _pytest.logging import LogCaptureFixture

from messaging.router import Router
from workspaces.capability.enums import CapabilityRequestState, ExecutionState
from workspaces.capability.helpers import CapabilitySequence
from workspaces.capability.schema import CapabilityExecution, CapabilityRequest
from workspaces.capability.services.capability_info import CapabilityInfo
from workspaces.capability.services.execution_manager import ExecutionManager
from workspaces.workflow.schema import WorkflowRequest
from workspaces.workflow.services.workflow_service import WorkflowService

pytest_plugins = ["testing.utils.conftest"]

logger = logging.getLogger("execution_manager")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


def reset_execution_mgr(ex_man: ExecutionManager) -> ExecutionManager:
    """
    Refresh this execution manager.

    :param ex_man: the execution manager
    :return: a brand-spanking-new exec manager
    """
    ex_man.queues = {}
    ex_man.execution_pool = []
    return ex_man


@pytest.fixture(scope="function")
def dummy_execution_manager(
    mock_capability_info: CapabilityInfo,
    mock_workflow_service: WorkflowService,
    test_router: Router,
) -> ExecutionManager:
    """
    Pytest fixture that instantiates a dummy ExecutionManager object for testing purposes

    :return: Dummy ExecutionManager
    """
    with patch("workspaces.capability.services.capability_queue.threading.Thread"):
        yield ExecutionManager(mock_capability_info, mock_workflow_service, test_router)


@pytest.fixture(scope="function")
def workflow_executing_message(
    mock_workflow_service: WorkflowService,
    mock_workflow_requests: List[WorkflowRequest],
) -> Dict:
    """
    Create a reusable workflow-executing message

    :return: Workflow-executing message
    """
    request = mock_workflow_requests[0]
    return {
        "service": "workflow",
        "routing_key": "workflow",
        "type": "workflow-executing",
        "subject": request.__json__(),
    }


@pytest.fixture(scope="function")
def workflow_complete_message(
    mock_workflow_service: WorkflowService, mock_workflow_requests: List[WorkflowRequest]
) -> Dict:
    """
    Create a reusable workflow-complete message

    :return: Workflow-complete message
    """
    request = mock_workflow_requests[0]
    return {
        "service": "workflow",
        "routing_key": "workflow",
        "type": "workflow-complete",
        "subject": request.__json__(),
    }


@pytest.fixture(scope="function")
def step_complete_message(mock_capability_execution: CapabilityExecution) -> Dict:
    """
    Build a capability "step-complete" message

    :param mock_capability_execution:
    :return:
    """
    return {
        "service": "capability",
        "routing_key": "capability",
        "subject": mock_capability_execution.__json__(),
        "type": "step-complete",
    }


def test_run_capability(
    dummy_execution_manager: ExecutionManager,
    mock_capability_requests: List[CapabilityRequest],
):
    """
    Tests that run_capability correctly creates an execution given a request
    and enqueues that execution if the first capability step is running a workflow
    """
    null_capability_request = mock_capability_requests[0]
    assert null_capability_request.state == CapabilityRequestState.Created.name
    # logger.info(f">>> the capability request: {null_capability_request}")
    null_execution = dummy_execution_manager.run_capability(null_capability_request)
    assert null_capability_request.state == CapabilityRequestState.Submitted.name
    assert null_execution.state == CapabilityRequestState.Created.name
    assert len(dummy_execution_manager.queues) == 1
    dummy_execution_manager = reset_execution_mgr(dummy_execution_manager)

    # Test with execution that will be placed into the execution pool rather than enqueued
    test_capability_request = mock_capability_requests[1]
    test_execution = dummy_execution_manager.run_capability(test_capability_request)
    assert len(dummy_execution_manager.execution_pool) == 1
    assert test_execution.state == ExecutionState.Waiting.name


def test_send_step_complete_workflow_executing(
    dummy_execution_manager: ExecutionManager,
    mock_capability_execution: CapabilityExecution,
    workflow_executing_message: Dict,
    step_complete_message: Dict,
):
    """
    Tests that sending a workflow-executing step complete message works properly
    """
    with patch("messaging.router.Router.receive_message"):
        dummy_execution_manager.send_step_complete(**workflow_executing_message)
        assert dummy_execution_manager.message_router.receive_message.called_with(
            step_complete_message
        )

    # Reset state of mock execution
    mock_capability_execution.current_step = 0


def test_send_step_complete_workflow_complete(
    dummy_execution_manager: ExecutionManager,
    mock_capability_execution: CapabilityExecution,
    workflow_complete_message: Dict,
    step_complete_message: Dict,
):
    """
    Tests that sending a workflow-complete step complete message works properly
    """
    with patch("messaging.router.Router.receive_message"):
        dummy_execution_manager.send_step_complete(**workflow_complete_message)
        assert dummy_execution_manager.message_router.receive_message.called_with(
            step_complete_message
        )

    # Reset state of mock execution
    mock_capability_execution.current_step = 0


def test_on_step_complete(
    dummy_execution_manager: ExecutionManager,
    mock_capability_execution: CapabilityExecution,
    step_complete_message: Dict,
):
    """
    Tests that on_step_complete correctly advances and completes executions
    """
    # Simulate starting conditions of queues
    dummy_execution_manager.enqueue_execution(mock_capability_execution)

    with patch("workspaces.capability.services.capability_info.CapabilityInfo.save_execution"):
        # _advance_execution will be called
        assert mock_capability_execution.current_step == 0
        dummy_execution_manager.on_step_complete(**step_complete_message)
        assert mock_capability_execution.current_step == 1
        dummy_execution_manager.capability_info.save_execution.assert_called_with(
            mock_capability_execution
        )
        # On AwaitWorkflow step; queues["null"].complete_execution will be called
        # On last step; _complete_execution will be called
        with patch(
            "workspaces.capability.services.capability_queue.CapabilityQueue.complete_execution"
        ):
            with patch("workspaces.capability.services.execution_manager.Router.send_message"):
                assert mock_capability_execution.on_last_step()
                dummy_execution_manager.on_step_complete(**step_complete_message)
                dummy_execution_manager.queues["null"].complete_execution.assert_called_with(
                    mock_capability_execution.id
                )
                dummy_execution_manager.capability_info.save_execution.assert_called_with(
                    mock_capability_execution
                )
                assert dummy_execution_manager.message_router.send_message.call_count == 1
                assert mock_capability_execution.state == "Complete"

    # Reset state of mock execution
    mock_capability_execution.current_step = 0


def test_on_step_complete_unrecognized(
    dummy_execution_manager: ExecutionManager,
    mock_capability_execution: CapabilityExecution,
    step_complete_message: Dict,
    caplog: LogCaptureFixture,
):
    """
    Tests that on_step_complete identifies a badly formatted message
    """
    step_complete_message["subject"]["type"] = "Unrecognizable"
    dummy_execution_manager.on_step_complete(**step_complete_message)
    messages = [record.message for record in caplog.records]
    assert "Unrecognized message type: Unrecognizable." in messages


def test_advance_execution(
    dummy_execution_manager: ExecutionManager,
    mock_capability_execution: CapabilityExecution,
):
    """
    Tests that an execution is properly advanced to its next step
    """
    step_sequence = CapabilitySequence.from_str(mock_capability_execution.steps)
    assert mock_capability_execution.current_step == 0
    dummy_execution_manager._advance_execution(mock_capability_execution, step_sequence)
    assert mock_capability_execution.current_step == 1
    assert mock_capability_execution.on_last_step()

    # Reset mock execution to have P&RW step as step 2
    mock_capability_execution.current_step = 0
    mock_capability_execution.steps = "await-qa\nprepare-and-run-workflow null\nawait-workflow"
    step_sequence = CapabilitySequence.from_str(mock_capability_execution.steps)
    dummy_execution_manager = reset_execution_mgr(dummy_execution_manager)

    # Test that prepare-and-run-workflow step is handled correctly
    dummy_execution_manager._advance_execution(mock_capability_execution, step_sequence)
    assert mock_capability_execution.current_step == 1
    # Assert that the execution has been queued for a workflow execution
    assert dummy_execution_manager.queues["null"].queue_manager.start.call_count == 1
    assert len(dummy_execution_manager.queues) == 1
    dummy_execution_manager._advance_execution(mock_capability_execution, step_sequence)
    assert mock_capability_execution.current_step == 2
    assert mock_capability_execution.on_last_step()

    # Reset state of mock execution
    mock_capability_execution.current_step = 0


def test_complete_execution(
    dummy_execution_manager: ExecutionManager, mock_capability_execution: CapabilityExecution
):
    """
    Tests that _complete_execution correctly marks an execution as complete and sends the
    corresponding message
    """
    with patch("messaging.router.Router.receive_message"):
        # Reset execution state
        mock_capability_execution.state = CapabilityRequestState.Created.name

        dummy_execution_manager._complete_execution(mock_capability_execution)
        assert dummy_execution_manager.message_router.receive_message.called_with(
            {
                "service": "capability",
                "routing_key": "capability",
                "subject": mock_capability_execution.__json__(),
                "type": "execution-complete",
            }
        )
        assert mock_capability_execution.state == "Complete"

    # Reset state of mock execution
    mock_capability_execution.current_step = 0


def test_enqueue_execution(
    dummy_execution_manager: ExecutionManager, mock_capability_execution: CapabilityExecution
):
    """
    Tests that executions are properly enqueued
    """
    dummy_execution_manager.execution_pool.append(mock_capability_execution)
    dummy_execution_manager.enqueue_execution(mock_capability_execution)
    assert len(dummy_execution_manager.queues) == 1
    assert len(dummy_execution_manager.queues["null"]) == 1
    assert dummy_execution_manager.queues["null"].queue_manager.start.call_count == 1
    assert mock_capability_execution not in dummy_execution_manager.execution_pool
